#! /usr/bin/env python

"""My first attempt at a pygame game. It implements the classic "falling-block
puzzle game." Most of this code is copied from the aliens.py example that comes
with pygame."""

from array import array
import whrandom
import os.path
import sys
import pygame
import pygame.image
from pygame.locals import *

try:
    import pygame.mixer
    pygame.mixer.pre_init(11025)
except:
    pygame.mixer = None

# constants

BLOCK_WIDTH     = 16    # px width of a falling block
FRAMES_PER_SEC  = 24    # gameplay fps
FRAMES_PER_TILE = 11    # number of frames before block falls 1 tile due to gravity
GRAVITY         = 1     # accel. due to gravity (pixels/frame)
JOY_THRESHOLD   = 0.4   # joystick deflection must exceed this value to move
KEY_REPEAT      = 6     # number of frames a key must be held before it repeats
PLAY_HEIGHT     = 13    # height of play area in blocks
PLAY_WIDTH      = 11    # width of play area in blocks
SCREEN_RECT     = Rect(0, 0, 320, 240)      # preferred window size
TERMINAL_VELOCITY = 12  # terminal velocity (pixels/frame)
WIN_STYLE       = DOUBLEBUF | FULLSCREEN    # preferred window style

# rectangle where play actually occurs

PLAY_RECT       = Rect(0, 0, (PLAY_WIDTH + 1) * BLOCK_WIDTH, (PLAY_HEIGHT + 1) * BLOCK_WIDTH)
PLAY_RECT.centery = SCREEN_RECT.centery
PLAY_RECT.left  = PLAY_RECT.top

# block data (quadruple of (x, y) offsets)

blocks = [
    # xx.
    # .xx
    # ...
    ((0, 0), (1, 0), (1, 1), (2, 1)),

    # .xx
    # xx.
    # ...
    ((1, 0), (2, 0), (0, 1), (1, 1)),

    # .x.
    # .x.
    # .xx
    ((1, 0), (1, 1), (1, 2), (2, 2)),

    # .x.
    # .x.
    # xx.
    ((1, 0), (1, 1), (0, 2), (1, 2)),

    # ...
    # xxx
    # .x.
    ((0, 1), (1, 1), (2, 1), (1, 2)),

    # .x..
    # .x..
    # .x..
    # .x..
    ((1, 0), (1, 1), (1, 2), (1, 3)),

    # xx
    # xx
    ((0, 0), (1, 0), (0, 1), (1, 1))
    ]

# 2d array of blocks that have fallen into play area
play_area = []

class DummySound:
    "A null/no-op sound. Used as a placeholder if a sound fails to load."
    def play(self):
        pass

class Particle:
    "A particle in the particle system."
    def __init__(self, image, pos, xvel, yvel):
        self.image = image
        self.rect = image.get_rect()
        self.rect.left = pos[0]
        self.rect.top = pos[1]
        self.xvel = xvel
        self.yvel = yvel

    def draw(self, screen):
        "Draws the particle onto the screen and moves it."
        screen.blit(self.image, self.rect)
        self.rect.top = self.rect.top + self.yvel
        self.rect.left = self.rect.left + self.xvel

        if self.yvel < TERMINAL_VELOCITY:
            self.yvel += GRAVITY

    def is_on_screen(self):
        "Returns true if the particle is on screen."
        return self.rect.top < SCREEN_RECT.height and self.rect.right >= 0 and self.rect.left < SCREEN_RECT.width
        

def block_height(b):
    "Returns the height of the block b, in tiles."
    return max(b[0][1], b[1][1], b[2][1], b[3][1])
    
def block_size(b):
    "Returns the largest width or height, in tiles, of the block specified in b."
    return max(b[0][0], b[0][1], b[1][0], b[1][1], b[2][0], b[2][1], b[3][0], b[3][1])

# This method returns the width and height of b in blocks. Unlike block_size,
# block_size_precise doesn't use (0, 0) as a reference point. Instead it returns
# the top left offset of the block in the tuple.
def block_size_precise(b):
    "Returns a width, height, offset-x, offset-y tuple of b in blocks."
    x = min(b[0][0], b[1][0], b[2][0], b[3][0])
    y = min(b[0][1], b[1][1], b[2][1], b[3][1])
    w = max(b[0][0], b[1][0], b[2][0], b[3][0]) - x + 1
    h = max(b[0][1], b[1][1], b[2][1], b[3][1]) - y + 1
    return (w, h, x, y)

def block_width(b):
    "Returns the width of the block b, in tiles."
    return max(b[0][0], b[1][0], b[2][0], b[3][0])

def can_move(x, y, b):
    "Test whether the block b can be positioned at (x,y)."
    if x + block_width(b) >= PLAY_WIDTH:
        return 0
    
    for u, v in b:
        if (y + v >= 0) and (x + u < 0 or x + u >= PLAY_WIDTH or y + v >= PLAY_HEIGHT or play_area[y + v][x + u] != 0):
            return 0
    return 1

def draw_number(screen, x, y, n):
    "Renders the number n, right-aligned at (x, y)."
    global digit_img
    
    arr = array('c', str(n))
    arr.reverse()
            
    for c in arr.tostring():
        x -= digit_img[int(c)].get_rect().width
        screen.blit(digit_img[int(c)], (x, y))
    
def load_image(file, alpha):
    "Loads an image and converts it to the display format, ready for play."
    file = os.path.join('data', file)
    try:
        surface = pygame.image.load(file)
    except pygame.error:
        raise SystemExit, 'Couldn\'t load image "%s" (%s)'%(file, pygame.get_error())
    if alpha:
        return surface.convert_alpha()
    else:
        return surface.convert()


def load_sound(file):
    "Loads a sound file."
    if not pygame.mixer:
        return DummySound()
    file = os.path.join('data', 'sounds', file)
    try:
        sound = pygame.mixer.Sound(file)
        return sound
    except pygame.error:
        print '*** Warning: Couldn\'t load sound "%s" (%s)'%(file, pygame.get_error())
    return DummySound()


def rot_block(b):
    "Rotates a block 90 deg. counter-clockwise, returning the rotated tuple."
    m = block_size(b)
    return ((b[0][1], m - b[0][0]),
            (b[1][1], m - b[1][0]),
            (b[2][1], m - b[2][0]),
            (b[3][1], m - b[3][0]))
    

last_tick = 0
ticks_per_frame = int((1.0 / FRAMES_PER_SEC) * 1000.0)

def wait_frame():
    "Inserts a delay to ensure the game runs at the correct speed."
    global last_tick, ticks_per_frame
    now = pygame.time.get_ticks()
    wait = ticks_per_frame - (now - last_tick)
    pygame.time.delay(wait)
    last_tick = pygame.time.get_ticks()

def main():
    global digit_img, FRAMES_PER_TILE, play_area
    
    # init SDL
    pygame.init()

    # set the display mode
    
    depth = pygame.display.mode_ok(SCREEN_RECT.size, WIN_STYLE, 32)
    screen = pygame.display.set_mode(SCREEN_RECT.size, WIN_STYLE, depth)

    pygame.mouse.set_visible(0)

    # initialize the joystick (if any)
    if pygame.joystick.get_init() and pygame.joystick.get_count():
        joy = pygame.joystick.Joystick(0)
        joy.init()
        if not joy.get_numaxes() >= 2 or not joy.get_numbuttons() >= 2:
            print '*** Warning: Joystick disabled-- not enough buttons/axes'
            joy.quit()
            joy = None
    else:
        joy = None

    # check that the mixer init'd

    if pygame.mixer and not pygame.mixer.get_init():
        pygame.mixer = None

    if not pygame.mixer:
        print '*** Warning: Sound disabled'

    # load the in-game resources

    # load the title screen and in-game branding images
    
    title_large = load_image('pygris-large.png', 1)

    title_large_rect = title_large.get_rect()
    title_large_rect.centerx = SCREEN_RECT.centerx
    title_large_rect.centery = SCREEN_RECT.centery
    
    title_small = load_image('pygris.png', 1)

    title_small_rect = title_small.get_rect()
    title_small_rect.right = SCREEN_RECT.right - 10
    title_small_rect.top = SCREEN_RECT.top + 10

    game_over = load_image('game-over.png', 1)

    game_over_rect = game_over.get_rect()
    game_over_rect.centerx = SCREEN_RECT.centerx
    game_over_rect.centery = SCREEN_RECT.centery

    instructions = load_image('instructions.png', 1)

    instructions_rect = instructions.get_rect()
    instructions_rect.centerx = SCREEN_RECT.centerx
    instructions_rect.centery = SCREEN_RECT.centery

    # load the coloured block images
    
    block_img = [None]  # A "blank" for block 0

    for filename in os.listdir(os.path.join('data', 'blocks')):
        block_img.append(load_image(os.path.join('blocks', filename), 1))

    # save a copy of the list of files in backgrounds

    background_filenames = []

    for filename in os.listdir(os.path.join('data', 'backgrounds')):
        background_filenames.append(os.path.join('backgrounds', filename))

    # load images for each digit

    digit_img = []

    for i in range(0, 10):
        digit_img.append(load_image(os.path.join('digits', str(i) + '.png'), 1))

    # load the score, lines and level images

    score_img = load_image('score.png', 1)
    level_img = load_image('level.png', 1)
    lines_img = load_image('lines.png', 1)
    
    level_rect = level_img.get_rect()
    level_rect.top = title_small_rect.bottom - 16
    level_rect.right = SCREEN_RECT.right - BLOCK_WIDTH

    score_rect = score_img.get_rect()
    score_rect.top = level_rect.centery + digit_img[0].get_rect().height
    score_rect.right = level_rect.right

    lines_rect = lines_img.get_rect()
    lines_rect.top = score_rect.centery + digit_img[0].get_rect().height
    lines_rect.right = level_rect.right

    # save a copy of the list of files in music

    music_filenames = []

    for filename in os.listdir(os.path.join('data', 'music')):
        music_filenames.append(os.path.join('data', 'music', filename))

    # load sounds

    score_sound = load_sound('bling.wav')
    block_sound = load_sound('tock.wav')
    game_over_sound = load_sound('wawawawa.wav')
    level_up_sound = load_sound('trumpets.wav')

    # loop until done

    done = 0
    game_done = 1

    while not done:
        # display title screen

        screen.blit(load_image(whrandom.choice(background_filenames), 0), (0, 0))
        screen.blit(title_large, title_large_rect)
        pygame.display.flip()

        # loop until keypress

        while 1:
            pygame.event.pump()
            keystate = pygame.key.get_pressed()
            if keystate[K_ESCAPE] or pygame.event.peek(QUIT):
                done = 1
                break
            elif keystate[K_SPACE] or joy and joy.get_button(0):
                game_done = 0
                break

        # display game instructions
        if not game_done:
            screen.blit(load_image(whrandom.choice(background_filenames), 0), (0, 0))
            screen.blit(instructions, instructions_rect)
            pygame.display.flip()

            # wait for keypress
            while 1:
                pygame.event.pump()
                keystate = pygame.key.get_pressed()
                if keystate[K_SPACE] or joy and joy.get_button(0):
                    break
                if keystate[K_ESCAPE]:
                    game_done = 1
                    break
            
        # set up game
        if not game_done:
            # load a random background image
            bg = load_image(whrandom.choice(background_filenames), 0)

            # create a random surface to frame the play area, scores, etc.

            play_surf = pygame.Surface((PLAY_RECT.width, PLAY_RECT.height))
            col = (whrandom.choice(range(255)), whrandom.choice(range(255)), whrandom.choice(range(255)))
            play_surf.fill(col)
            play_surf.set_alpha(128)

            # draw faded out play area
            bg.blit(play_surf, PLAY_RECT)

            # draw faded-out "next block" area in lower-right

            play_surf = pygame.Surface((BLOCK_WIDTH * 5, BLOCK_WIDTH * 5))
            play_surf.fill(col)
            play_surf.set_alpha(128)

            next_rect = Rect(0, 0, BLOCK_WIDTH * 5, BLOCK_WIDTH * 5)
            next_rect.bottom = SCREEN_RECT.bottom - BLOCK_WIDTH / 2
            next_rect.left = PLAY_RECT.right + (SCREEN_RECT.width - (PLAY_RECT.right + next_rect.width)) / 2

            bg.blit(play_surf, next_rect)

            play_surf = None

            # init in-game vars

            level = 1           # start at level 1
            nlines = 0          # no lines
            score = 0           # points for lines
            
            left_pressed = 0    # keyboard/joystick input vars-- nothing pressed
            right_pressed = 0
            down_pressed = 0
            drop_pressed = 0
            rotate_pressed = 0

            particles = []
 
            # choose current and next block, randomize coloured pieces

            cur_block = whrandom.randint(0, len(blocks) - 1)
            cur_block_tiles = [whrandom.randint(1, len(block_img) - 1) for _ in range(4)]
            cur_block_coords = blocks[cur_block]

            next_block = whrandom.randint(0, len(blocks) - 1)
            next_block_tiles = [whrandom.randint(1, len(block_img) - 1) for _ in range(4)]
            next_block_coords = blocks[next_block]

            # init block position

            cur_block_x = (PLAY_WIDTH - block_width(cur_block_coords)) / 2
            cur_block_y = -block_height(cur_block_coords)
            fall_counter = 0

            # initialize play area, 2d array to all zeros (= empty)

            play_area = []

            for i in range(PLAY_HEIGHT):
                play_area.append([0 for _ in range(PLAY_WIDTH)])

            # start music playing!
            try:
                pygame.mixer.music.load(whrandom.choice(music_filenames))
                pygame.mixer.music.set_volume(0.7)
                pygame.mixer.music.play(-1)
            except:
                pass

        # game loop
        while not game_done:
            # clear screen
            screen.blit(bg, (0, 0))

            # draw play area

            for v in range(PLAY_HEIGHT):
                for u in range(PLAY_WIDTH):
                    if play_area[v][u]:
                        screen.blit(block_img[play_area[v][u]], (PLAY_RECT.left + (BLOCK_WIDTH / 2) + BLOCK_WIDTH * u, PLAY_RECT.top + (BLOCK_WIDTH / 2) + BLOCK_WIDTH * v))

            # draw current block

            for i in range(4):
                x, y = cur_block_coords[i]

                if cur_block_y + y >= 0:
                    screen.blit(block_img[cur_block_tiles[i]], (PLAY_RECT.left + (BLOCK_WIDTH / 2) + (cur_block_x + x) * BLOCK_WIDTH,
                                                                PLAY_RECT.top + (BLOCK_WIDTH / 2) + (cur_block_y + y) * BLOCK_WIDTH))

            # draw next block (centered, in "next block" area)

            (w, h, u, v) = block_size_precise(next_block_coords)
            
            u = ((5 - (w + 2 * u)) * BLOCK_WIDTH / 2) + next_rect.left;
            v = ((5 - (h + 2 * v)) * BLOCK_WIDTH / 2) + next_rect.top;

            for i in range(4):
                x, y = next_block_coords[i]
                screen.blit(block_img[next_block_tiles[i]], (u + x * BLOCK_WIDTH,
                                                             v + y * BLOCK_WIDTH))

            # draw particles (if any)

            for p in particles:
                p.draw(screen)

            # remove any particles that are off screen
            particles = filter(lambda p: p.is_on_screen(), particles)

            # draw level, score and lines

            draw_number(screen, level_rect.right, level_rect.centery, level)
            draw_number(screen, score_rect.right, score_rect.centery, score)
            draw_number(screen, lines_rect.right, lines_rect.centery, nlines)

            # draw level, score and lines titles

            screen.blit(level_img, level_rect)
            screen.blit(score_img, score_rect)
            screen.blit(lines_img, lines_rect)

            # draw title
            screen.blit(title_small, title_small_rect)

            # read input

            pygame.event.pump()
            keystate = pygame.key.get_pressed()

            if not keystate[K_LEFT] and not (joy and joy.get_axis(0) < -JOY_THRESHOLD):
                left_pressed = 0
            else:
                left_pressed += 1

            if not keystate[K_RIGHT] and not (joy and joy.get_axis(0) > JOY_THRESHOLD):
                right_pressed = 0
            else:
                right_pressed += 1

            if not keystate[K_DOWN] and not (joy and joy.get_axis(1) > JOY_THRESHOLD):
                down_pressed = 0
            else:
                down_pressed += 1

            if not keystate[K_LALT] and not keystate[K_RALT] and not (joy and joy.get_button(1)):
                drop_pressed = 0
            else:
                drop_pressed += 1
                    

            if not keystate[K_SPACE] and not (joy and joy.get_button(0)):
                rotate_pressed = 0
            else:
                rotate_pressed += 1

            if keystate[K_ESCAPE] or pygame.event.peek(QUIT):
                # stop the music
                try:
                    pygame.mixer.music.stop()
                except:
                    pass
                
                game_done = 1
            elif left_pressed % KEY_REPEAT == 1 and can_move(cur_block_x - 1, cur_block_y, cur_block_coords):
                cur_block_x -= 1
            elif right_pressed % KEY_REPEAT == 1 and can_move(cur_block_x + 1, cur_block_y, cur_block_coords):
                cur_block_x += 1
            elif rotate_pressed % KEY_REPEAT == 1 and can_move(cur_block_x, cur_block_y, rot_block(cur_block_coords)):
                cur_block_coords = rot_block(cur_block_coords)
            elif down_pressed % KEY_REPEAT == 1 and can_move(cur_block_x, cur_block_y + 1, cur_block_coords):
                cur_block_y += 1
                fall_counter = 0
            elif drop_pressed == 1:
                while can_move(cur_block_x, cur_block_y + 1, cur_block_coords):
                    cur_block_y += 1
                fall_counter = FRAMES_PER_TILE # "lock" immediately

            # fall counter makes the block fall slower than the frame rate
            fall_counter += 1

            # time to fall down due to gravity (if possible)
            if fall_counter >= FRAMES_PER_TILE and can_move(cur_block_x, cur_block_y + 1, cur_block_coords):
                cur_block_y += 1
                fall_counter = 0

            # If the block has hit the bottom, refresh in-game vars, etc. The
            # fall_counter < FRAMES_PER_TILE may seem strange-- this is so the
            # player can have time to "slide" a block under another one. Without
            # it the block would touch the bottom and immediately "freeze."
            elif fall_counter >= FRAMES_PER_TILE and not can_move(cur_block_x, cur_block_y + 1, cur_block_coords):
                # play 'block placed' sound
                block_sound.play()
                
                # write tiles into play_area
                for i in range(4):
                    x = cur_block_x + cur_block_coords[i][0]
                    y = cur_block_y + cur_block_coords[i][1]

                    if x >= 0 and x < PLAY_WIDTH and y >= 0 and y < PLAY_HEIGHT:
                        play_area[y][x] = cur_block_tiles[i]

                # move to next block
                
                cur_block = next_block
                cur_block_tiles = next_block_tiles
                cur_block_coords = next_block_coords

                # reset next block
                
                next_block = whrandom.randint(0, len(blocks) - 1)
                next_block_tiles = [whrandom.randint(1, len(block_img) - 1) for _ in range(4)]
                next_block_coords = blocks[next_block]

                # init block position

                cur_block_x = (PLAY_WIDTH - block_width(cur_block_coords)) / 2
                cur_block_y = -block_height(cur_block_coords)
                fall_counter = 0

                # check for completed lines

                lines_this_time = 0

                for i in range(PLAY_HEIGHT):
                    # if line contains no zeros, it is full
                    if not play_area[i].count(0):
                        # create pretty block particles
                        for j in range(PLAY_WIDTH):
                            particles.append(Particle(block_img[play_area[i][j]], (j * BLOCK_WIDTH + BLOCK_WIDTH / 2 + PLAY_RECT.left,
                                                                                   i * BLOCK_WIDTH + BLOCK_WIDTH / 2 + PLAY_RECT.top), whrandom.randint(-10, 10), whrandom.randint(-10, -4)))
                        
                        # remove full line
                        del play_area[i]

                        # insert new, empty line at top
                        play_area.insert(0, [0 for _ in range(PLAY_WIDTH)])

                        # loop must re-check current line
                        i -= 1

                        # increment score
                        lines_this_time += 1
                        nlines += 1

                if lines_this_time > 0:
                    # make old lines & score fly off

                    x = score_rect.right
                    y = score_rect.centery
                    arr = array('c', str(score))
                    arr.reverse()

                    for c in arr.tostring():
                        x -= digit_img[int(c)].get_rect().width
                        particles.append(Particle(digit_img[int(c)], (x, y), whrandom.randint(-10, 10), whrandom.randint(-10, -2)))

                    x = lines_rect.right
                    y = lines_rect.centery
                    arr = array('c', str(nlines))
                    arr.reverse()

                    for c in arr.tostring():
                        x -= digit_img[int(c)].get_rect().width
                        particles.append(Particle(digit_img[int(c)], (x, y), whrandom.randint(-10, 10), whrandom.randint(-10, -2)))

                    # add to score
                    score += pow(3, lines_this_time - 1)

                    # play "bling" sound for scoring
                    score_sound.play(lines_this_time - 1)

                    # increase level
                    if nlines > pow(2, 5 + level) and level < 5:
                        # make old level fly off

                        x = level_rect.right
                        y = level_rect.centery
                        arr = array('c', str(level))
                        arr.reverse()

                        for c in arr.tostring():
                            x -= digit_img[int(c)].get_rect().width
                            particles.append(Particle(digit_img[int(c)], (x, y), whrandom.randint(-10, 10), whrandom.randint(-10, -2)))

                        # level goes up one                        
                        level += 1

                        # make the blocks fall faster
                        FRAMES_PER_TILE -= 2
                        
                        # play some different music
                        try:
                            pygame.mixer.music.load(whrandom.choice(music_filenames))
                            pygame.mixer.music.set_volume(0.7)
                            pygame.mixer.music.play(-1)
                        except:
                            pass

                        # play the "level up" sound
                        level_up_sound.play()
                        
                # check whether next block can be placed
                if not can_move(cur_block_x, cur_block_y, cur_block_coords):
                    # draw game over message
                    screen.blit(game_over, game_over_rect)

                    # draw level, score and lines

                    draw_number(screen, level_rect.right, level_rect.centery, level)
                    draw_number(screen, score_rect.right, score_rect.centery, score)
                    draw_number(screen, lines_rect.right, lines_rect.centery, nlines)

                    # draw level, score and lines titles

                    screen.blit(level_img, level_rect)
                    screen.blit(score_img, score_rect)
                    screen.blit(lines_img, lines_rect)                    

                    # fade out the music
                    try:
                        pygame.mixer.music.fadeout(1000)
                    except:
                        pass
                    
                    game_over_sound.play()
                    game_done = 1

            pygame.display.flip()

            if game_done and not keystate[K_ESCAPE]:
                while not (keystate[K_ESCAPE] or joy and joy.get_button(0)):
                    pygame.event.pump()
                    keystate = pygame.key.get_pressed()
            else:
                # keep the game running at a set speed
                wait_frame()

#if python says run, let's run!
if __name__ == '__main__':
    main()
